/* 
**  mod_dnsbl.c -- Apache dnsbl module
**
**  this module constructs a DNS query from the URI and asks the DNS
**  for a classification along the lines of the well known DNS blacklists
**  against spam mail. It then checks the actions table and accepts or
**  rejects the request based on the result of this
**
**  (c) Dr. Andreas Mueller, Beratung und Entwicklung
**
**  $Id: mod_dnsbl.c,v 1.17 2008/01/05 15:55:41 afm Exp $
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_log.h"
#include "ap_config.h"
#include "ap_compat.h"
#include <http_core.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <ctype.h>

#define APLOG_STATUS    0,
#include <apr.h>
#include <apr_strings.h>
#include <apr_base64.h>
#include <apr_hooks.h>
#include <apr_tables.h>
#include <apr_pools.h>

#define	DNSBL_PASS	0
#define	DNSBL_BLOCK	1
#define	DNSBL_SKIPAUTH	2

module MODULE_VAR_EXPORT dnsbl_module;

/* configuration record */
typedef struct dnsbl_config_rec {
	apr_array_header_t	*suffix;
	char		*template;
	char		*message;
	char		*contact;
	apr_table_t	*actions;
	apr_table_t	*reasons;
	int		defaultaction;
	int		authoritative;
	int		recurse;
} dnsbl_config_rec;

static dnsbl_config_rec	*dnsbl_getconfig(request_rec *r) {
    server_rec		*s;
    dnsbl_config_rec	*rec;
    if (r == NULL) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r, "NULL request???");
	return NULL;
    }
    s = r->server;
    if (s == NULL) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r, "NULL server???");
    }
    rec = (dnsbl_config_rec *)ap_get_module_config(s->module_config,
	&dnsbl_module);
    if (rec == NULL) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r,
	    "NULL config record???");
    }
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"found config rec");
    return rec;
}

/* function declarations (only what is necessary) */
static int	dnsbl_check_name(request_rec *r, const char *suffix,
	const char *name);

/* convenience functions to parse an action string */
static time_t	dnsbl_parse_time(const char *t) {
    struct tm	st;
    time_t	now = time(NULL);
    localtime_r(&now, &st);
    st.tm_sec = 0;
    sscanf(t, "%02d", &st.tm_hour);
    sscanf(t + 3, "%02d", &st.tm_min);
    return mktime(&st);
}

static int	dnsbl_check_time(request_rec *r, const char *t, int l) {
    time_t		start, now, end;

    /* ignore time specs of incorrect length */
    if (l != 11) {
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	    "[%d] incorrect time spec (wrong length %d): %*.*s", (int)getpid(),
	    l, l, l, t);
	return 0;
    }

    /* split into start and stop time */
    start = dnsbl_parse_time(t);
    end = dnsbl_parse_time(t + 6);

    /* check whether now is in the interval */
    now = time(NULL);
    if ((start < now) && (now < end)) {
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	    "[%d] time spec allows access", (int)getpid());
	return 1;
    }
    return 0;
}

static int	dnsbl_check_singleaction(request_rec *r, const char *action,
	int length, const char *item) {
    int	l;

    /* log action */
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	    "[%d] check action %*.*s for %s", (int)getpid(),
	    length, length, action, item);

    /* if the action does not begin with the item string, forget it */
    l = strlen(item);
    if (strncmp(action, item, l))
	return 0;

    /* check for a time specification */
    if (*(action + l) == '/') {
	return dnsbl_check_time(r, action + l + 1, length - l - 1);
    }
    return 1; /* no time qualifier, match already established */
}

static int	dnsbl_check_actionstring(request_rec *r, const char *action,
	const char *item) {
    int	rc;
    const char	*p, *q;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] checking actionstring %s against item %s", (int)getpid(),
	action, item);

    /* start looking for items 					*/
    p = action;
    while (p) {
	int	l = 0;
	/* find the end of the current item */
	q = strchr(p, ',');
	if (NULL == q) {
	    l = strlen(p);
	} else {
	    l = q - p;
	}
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	   "[%d] checking component %*.*s", (int)getpid(), l, l, p);

	/* check the action */
	rc = dnsbl_check_singleaction(r, p, l, item);
	if (rc > 0)
	    return rc;

	/* go to next component */
	if (q)
		p = q + 1;
	else
		p = q;
    }

    /* no such item found						*/
    return 0;
}


/* display the error page stuff */
static char	*template =
	"<html>\n"
	"<head>\n"
	"<title>Page blocked</title>\n"
	"</head>\n"
	"<body>\n"
	"<h1>Page blocked by mod_dnsbl</h1>\n"
	"<p>\n"
	"The page\n"
	"<blockquote>\n"
	"%u" /* URI */
	"</blockquote>\n"
	"was blocked by mod_dnsbl for the following reason:\n"
	"<blockquote>\n"
	"%r" /* Reason for block */
	"</blockquote>\n"
	"Please contact your system administrator at \n"
	"<a href=\"mailto:%c\">%c</a> if you don't agree with this decision.\n"
	"</p>\n"
	"</body>\n"
	"</html>\n";

/* The sample content handler */
static void	dnsbl_handler(request_rec *r, const char *reason) {
    int			f;
    dnsbl_config_rec	*rec = dnsbl_getconfig(r);
    struct stat		sb;
    size_t		l;
    char		*uri = "(unknown)";
    const char		*p = NULL;
    char		infopage[HUGE_STRING_LEN];
    int			i, j;

    /* we must somehow find the URI and the reason from the request */
    uri = r->uri;

    /* check for template */
    if ((NULL == rec->template) && (NULL == rec->message))
	goto havetemplate;

    /* if we have a message string, use that one */
    if (NULL != rec->message) {
	template = rec->message;
	goto havetemplate;
    }

    /* open the template for the reason page */
    if (stat(rec->template, &sb) < 0) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r,
	    "[%d] problem reading template %s", (int)getpid(), rec->template);
	goto havetemplate;
    }
    l = sb.st_size + 1;
    if ((f = open(rec->template, O_RDONLY)) < 0) {
 	ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r, "[%d] no template "
	    "found, using default", (int)getpid());
    }

    /* read the contents of the template */
    template = (char *)alloca(l);
    memset(template, 0, l);
    if (sb.st_size != read(f, template, sb.st_size)) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r,
	    "[%d] failed to read template", (int)getpid());
    }
    close(f);

    /* insert the info about the page */
havetemplate:
    ap_log_rerror(APLOG_MARK, APLOG_INFO, APLOG_STATUS r, "uri %s blocked: %s",
	uri, reason);

    /* the output headers should have cache control */
    apr_table_setn(r->headers_out, "Cache-Control", "no-cache");
    apr_table_setn(r->headers_out, "Pragma", "no-cache");

    /* format block reason */
    i = j = 0;
    while (i < sizeof(infopage) - 1 && template[j]) {
	if (template[j] != '%') {
	    infopage[i++] = template[j++];
	} else {
	    j++;
	    switch (template[j]) {
	    case 'u':	p = uri; j++;
			break;
	    case 'r':	p = reason; j++;
			break;
	    case 'c':	p = rec->contact; j++;
			break;
	    case '\0':	p = NULL;
			break;
	    default:	infopage[i++] = template[j++];
			break;
	    }
	    if (p) {
		while (i < sizeof(infopage) - 1 && *p) {
		    infopage[i++] = *p++;
		}
	    }
	}
    }
    infopage[i] = '\0';

    /* send this stuff as an error page to the client */
    ap_custom_response(r, HTTP_FORBIDDEN, infopage);
}

/* check whether a character is a legal DNS character */
static int	dnsbl_legal(char c) {
    if (isalnum(c)) return 1;
    if (c == '-') return 1;
    return 0;
}

/* convert a path component into a legal DNS name */

/* check a url path */
static int	dnsbl_check_path(request_rec *r, const char *suffix) {
    dnsbl_config_rec	*rec;
    char		*path;
    char		*oldname, *newname;
    char		*p, *q;
    rec = dnsbl_getconfig(r);
    int			level, hostl, pathl, sl, rc;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] perform URL path analysis for %s", (int)getpid(), r->uri);

    /* build a working copy of the URL path */
    pathl = strlen(r->parsed_uri.path);
    path = (char *)alloca(pathl + 1);
    strcpy(path, r->parsed_uri.path + 1);

    /* remove (replace by -) all illegal characters from the path */
    p = path;
    while (*p) {
	if (!dnsbl_legal(*p)) {
		*p = '-';
	}
	p++;
    }

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] working on path %s", (int)getpid(), path);

    /* prepare two working buffers for building the path */
    hostl = strlen(r->parsed_uri.hostname);
    sl = strlen(suffix);
    oldname = (char *)alloca(hostl + pathl + sl + 9);
    newname = (char *)alloca(hostl + pathl + sl + 9);
    strcpy(oldname, r->parsed_uri.hostname);
    strcat(oldname, ".");
    strcat(oldname, suffix);

    /* start with the hostname, and add new components from the path	*/
    level = 1;
    p = q = path;
    do {
	/* find the next component in the path string */
	q = strchr(p, '/');
	if (q) {
	    *q = '\0';
	    q++;
	} else {
	    q = p + strlen(p);
	}

	/* add the path component to the name */
	sprintf(newname, "%s.%s", p, oldname);
	rc = dnsbl_check_name(r, suffix, newname);
	if (rc >= 0) {
	    return rc;
	}
	p = q;
	level++;
    } while ((level < rec->recurse) && (*p));

    /* if we get here, there was no match */
    return DECLINED;
}

/* check a name */
static int	dnsbl_check_name(request_rec *r, const char *suffix,
	const char *name) {
    dnsbl_config_rec	*rec;
    char		*addr;
    const char		*action, *reason;
    struct addrinfo	*ai = NULL, *aip;
    int			rc = -1;

    rec = dnsbl_getconfig(r);

    /* perform a DNS lookup for this host */
    if (getaddrinfo(name, NULL, NULL, &ai) != 0) {
	/* if the name is not found, then we don't care			*/
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	    "[%d] name %s not found", (int)getpid(), name);
	return rc;
    }
    if (NULL == ai) {
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	    "[%d] name %s has no addresses", (int)getpid(), name);
	return rc;
    }

    /* check whether any of the addresses is in the table */
    for (aip = ai; NULL != aip; aip = aip->ai_next) {
	/* we should only get IPv4 addresses, so we just skip others	*/
	if (aip->ai_family != PF_INET)
	    continue;

	/* convert the current address to a string			*/
	addr = inet_ntoa(((struct sockaddr_in *)aip->ai_addr)->sin_addr);

	/* check for the address that triggers recursion		*/
	if (0 == strcmp(addr, "127.255.255.255")) {
	    if (rec->recurse > 0) {
		rc = dnsbl_check_path(r, suffix);
		if (rc >= 0) {
		    goto cleanup;
		}
	    }
	    continue;
	}

	/* look up the address in the rule tables			*/
	action = apr_table_get(rec->actions, addr);
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	    "[%d] %s has class ip %s, action = %s", (int)getpid(), name,
	    addr, (action) ? action : "not found");

	/* don't do anything if we have no action for this address	*/
	if (action == NULL)
	     continue;

	/* legal actions so far are block and pass			*/
	apr_table_set(r->notes, "dnsbl:action", action);
	if (dnsbl_check_actionstring(r, action, "noscan")) {
	    apr_table_set(r->notes, "dnsbl:scan", "no");
	}
	if (dnsbl_check_actionstring(r, action, "scan")) {
	    apr_table_set(r->notes, "dnsbl:scan", "yes");
	}
	if (dnsbl_check_actionstring(r, action, "block")) {
		/* output reason page */
		reason = apr_table_get(rec->reasons, addr);
		dnsbl_handler(r, (reason) ? reason : addr);
		rc = HTTP_FORBIDDEN;
		goto cleanup;
	}
	if (dnsbl_check_actionstring(r, action, "pass")) {
		rc = OK;	/* grant access */
		goto cleanup;
	}
	if (dnsbl_check_actionstring(r, action, "skipauth")) {
		apr_table_set(r->notes, "dnsbl:skipauth", "yes");
		rc = OK;	/* skip authentication */
		goto cleanup;
	}
    }

cleanup:
    if (ai)
	freeaddrinfo(ai);
    return rc;
}

/* check an individual suffix */
static int	dnsbl_check_suffix(request_rec *r, const char *suffix) {
    dnsbl_config_rec	*rec;
    char		newname[1024];

    rec = dnsbl_getconfig(r);

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] checking suffix %s", (int)getpid(), suffix);

    /* only check non-proxy requests */
    if (0 == r->proxyreq) {
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, APLOG_STATUS r,
	    "[%d] bypassing non-proxy request", (int)getpid());
	return OK;
    }

    /* build the new hostname from the suffix				*/
    if (r->parsed_uri.hostname == NULL)
	return DECLINED;
    snprintf(newname, sizeof(newname), "%s.%s", r->parsed_uri.hostname,
	suffix);
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] checking name %s", (int)getpid(), newname);

    /* check this name */
    return dnsbl_check_name(r, suffix, newname);
}

/* access checker */
static int	dnsbl_check(request_rec *r) {
    char		**suffixes;
    dnsbl_config_rec	*rec;
    int			i, rc = OK;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] check request for %s", (int)getpid(), r->uri);

    /* get the configuration record */
    rec = dnsbl_getconfig(r);
    if (NULL == rec) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r,
	    "[%d] no server record", (int)getpid());
	return DECLINED;
    }

    /* iterate through the domain suffixes */
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] check %d suffixes", (int)getpid(), rec->suffix->nelts);
    suffixes = (char **)(rec->suffix->elts);
    for (i = 0; i < rec->suffix->nelts; i++) {
	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	    "[%d] check suffix %s", (int)getpid(), suffixes[i]);
	rc = dnsbl_check_suffix(r, suffixes[i]);
	/* if rc!=-1 we have found the name and have a decision */
	if (rc != -1) {
	    return rc;
	}
    }
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] no rule activated: fall through to default", (int)getpid());

    /* if we get here, then the name was not found, so we should perform
       the default action. Compute the return code for the default action */
    rc = (rec->defaultaction == DNSBL_PASS) ? DECLINED : HTTP_FORBIDDEN;
    if (rc == HTTP_FORBIDDEN) {
	dnsbl_handler(r, "default action");
    }
    return rc;
}

/* auth/authz checker */
static int	dnsbl_auth(request_rec *r) {
    dnsbl_config_rec	*rec;
    const char		*action;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] auth check for %s", (int)getpid(), r->uri);
    rec = dnsbl_getconfig(r);
    if (!rec->authoritative)
	return DECLINED;
    action = apr_table_get(r->notes, "dnsbl:skipauth");
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS r,
	"[%d] found action %s", (int)getpid(), action);
    if (NULL == action)
	return DECLINED;
    if (0 == strcmp(action, "yes"))
	return OK;
    return DECLINED;
}

/* procedures related to configuration */
static const char	*dnsbl_add_action(cmd_parms *parms, void *mconfig,
    const char *address, const char *action, const char *reason) {
    dnsbl_config_rec	*rec = (dnsbl_config_rec *)ap_get_module_config(
	parms->server->module_config, &dnsbl_module);

    ap_log_error(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS 0,
	"adding action %s -> %s (%s)", address, action, reason);
    /* previously, we did some validation of the action string, but	*/
    /* with the additional option, this has become a little unwieldy	*/

    /* store the address and the action pair in the actions table	*/
    if (rec->actions == NULL) {
	ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, APLOG_STATUS 0,
	    "no actions table");
	return "no actions table for configuration";
    }
    apr_table_set(rec->actions, address, action);
    if (reason)
	apr_table_set(rec->reasons, address, reason);
    return NULL;
}

/* configuration methods */
static const char	*dnsbl_default_action_slot(cmd_parms *parms,
    void *mconfig, const char *action) {
    dnsbl_config_rec	*rec = (dnsbl_config_rec *)ap_get_module_config(
	parms->server->module_config, &dnsbl_module);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS 0,
	"setting default action to %s", action);
    if (0 == strcmp(action, "pass")) {
	rec->defaultaction = DNSBL_PASS;
	return NULL;
    }
    if (0 == strcmp(action, "block")) {
	rec->defaultaction = DNSBL_BLOCK;
	return NULL;
    }
    return "unknown default action";
}

static const char	*dnsbl_template_slot(cmd_parms *parms,
    void *mconfig, const char *value) {
    struct stat	sb;
    dnsbl_config_rec	*rec = (dnsbl_config_rec *)ap_get_module_config(
	parms->server->module_config, &dnsbl_module);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS 0,
	"setting template to %s", value);
    rec->template = apr_pstrdup(parms->pool, value);
    if (stat(rec->template, &sb) < 0) {
	ap_log_error(APLOG_MARK, APLOG_ERR, APLOG_STATUS 0,
	    "template file %s not found", rec->template);
	return "template file not found";
    }
    return NULL;
}

static const char	*dnsbl_contact_slot(cmd_parms *parms,
    void *mconfig, const char *value) {
    dnsbl_config_rec	*rec = (dnsbl_config_rec *)ap_get_module_config(
	parms->server->module_config, &dnsbl_module);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS 0,
	"setting contact address to %s", value);
    rec->contact = apr_pstrdup(parms->pool, value);
    return NULL;
}

static const char	*dnsbl_suffix_slot(cmd_parms *parms,
    void *mconfig, const char *value) {
    dnsbl_config_rec	*rec = (dnsbl_config_rec *)ap_get_module_config(
	parms->server->module_config, &dnsbl_module);
    char	**newsuffix;
    ap_log_error(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS 0,
	"adding suffix %s", value);
    newsuffix = (char **)apr_array_push(rec->suffix);
    *newsuffix = apr_pstrdup(parms->pool, value);
    return NULL;
}

static const char	*dnsbl_set_flag_slot(cmd_parms *parms, void *ptr,
	int flag) {
    dnsbl_config_rec	*rec = (dnsbl_config_rec *)ap_get_module_config(
	parms->server->module_config, &dnsbl_module);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS 0,
	"setting authoritative flag to %d", flag);
    rec->authoritative = flag;
    return NULL;
}

static const char	*dnsbl_set_recursion_slot(cmd_parms *parms, void *ptr,
	const char *recurse_string) {
    dnsbl_config_rec	*rec = (dnsbl_config_rec *)ap_get_module_config(
	parms->server->module_config, &dnsbl_module);
    char	*recurse_end;
    long	recurse = strtol(recurse_string, &recurse_end, 10);
    ap_log_error(APLOG_MARK, APLOG_DEBUG, APLOG_STATUS 0,
	"setting recursion depth to %ld", recurse);
    rec->recurse = recurse;
    return NULL;
}

static const command_rec	dnsbl_cmds[] = {
    AP_INIT_TAKE1(
	"DnsblTemplate", dnsbl_template_slot,
	(void *)APR_OFFSETOF(dnsbl_config_rec, template),
	OR_ALL,
	"Template file to be displayed if the remote system is not blocked."
    ),
    AP_INIT_ITERATE(
	"DnsblSuffix", dnsbl_suffix_slot,
	(void *)APR_OFFSETOF(dnsbl_config_rec, suffix),
	OR_ALL,
	"DNS suffixes within which to search for hosts"
    ),
    AP_INIT_TAKE1(
	"DnsblContact", dnsbl_contact_slot,
	(void *)APR_OFFSETOF(dnsbl_config_rec, contact),
	OR_ALL,
	"Contact email address for block page"
    ),
    AP_INIT_TAKE1(
	"DnsblDefaultAction", dnsbl_default_action_slot,
	(void *)APR_OFFSETOF(dnsbl_config_rec, defaultaction),
	OR_ALL,
	"Set default action"
    ),
    AP_INIT_TAKE23(
	"DnsblAction", dnsbl_add_action,
	(void *)APR_OFFSETOF(dnsbl_config_rec, actions),
	OR_ALL,
	"Define how to handle a given content class"
    ),
    AP_INIT_FLAG(
	"DnsblAuthoritative", dnsbl_set_flag_slot,
	(void *)APR_OFFSETOF(dnsbl_config_rec, authoritative),
	OR_ALL,
	"Set whether dnsbl should be allowed to override authentication"
    ),
    AP_INIT_TAKE1(
	"DnsblRecursionDepth", dnsbl_set_recursion_slot,
	(void *)APR_OFFSETOF(dnsbl_config_rec, recurse),
	OR_ALL,
	"Set recursion depth for URL path analysis"
    ),
    AP_INIT_TAKE1(
	"DnsblMessage", ap_set_string_slot,
	(void *)APR_OFFSETOF(dnsbl_config_rec, message),
	OR_ALL,
	"Set message format string"
    ),
    { NULL }
};

static void	*dnsbl_create_server_config(apr_pool_t *p, server_rec *rec) {
    dnsbl_config_rec	*cfg = (dnsbl_config_rec *)apr_pcalloc(p,
					sizeof(dnsbl_config_rec));
    cfg->template = NULL;
    cfg->contact = "root@localhost";
    cfg->actions = apr_table_make(p, 0);
    cfg->reasons = apr_table_make(p, 0);
    cfg->suffix = apr_array_make(p, 3, sizeof(char *));
    cfg->defaultaction = DNSBL_PASS;
    cfg->authoritative = 0;
    cfg->recurse = 0;
    cfg->message = NULL;
    return cfg;
}

static void	*dnsbl_merge_server_config(apr_pool_t *p, void *pp, void *cp) {
    dnsbl_config_rec	*parent = (dnsbl_config_rec *)pp;
    dnsbl_config_rec	*child = (dnsbl_config_rec *)cp;
    dnsbl_config_rec	*merged = (dnsbl_config_rec *)apr_pcalloc(p,
					sizeof(dnsbl_config_rec));
    /* merge string valued fields */
    merged->template = (child->template) ? child->template : parent->template;
    merged->message = (child->message) ? child->message : parent->message;
    merged->suffix = (child->suffix) ? child->suffix : parent->suffix;
    merged->contact = (child->contact) ? child->contact : parent->contact;

    /* merge the parent table with the child table */
    merged->actions = apr_table_overlay(p, parent->actions, child->actions);
    merged->reasons = apr_table_overlay(p, parent->reasons, child->reasons);
    return merged;
}

/* Dispatch list for API hooks */
static void	dnsbl_register_hooks(apr_pool_t *p) {
    static const char	*aszPre[] = { NULL };
    static const char	*aszSucc[] = { NULL };
    ap_hook_access_checker(dnsbl_check, aszPre, aszSucc, APR_HOOK_MIDDLE);
    ap_hook_check_user_id(dnsbl_auth, aszPre, aszSucc, APR_HOOK_FIRST);
    ap_hook_auth_checker(dnsbl_auth, aszPre, aszSucc, APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA	dnsbl_module = {
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    dnsbl_create_server_config,
    dnsbl_merge_server_config,
    dnsbl_cmds,
    dnsbl_register_hooks
};

